//********************************************************************
//*** Thread/CondVar.cc
//*** Copyright (c) 2001-2005 by Markus Winand <mws@fatalmind.com>
//*** $Id: CondVar.cc,v 1.10 2005/03/28 09:10:21 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of 
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#include "Thread/LockFwd.hh"
#include "Thread/CondVar.hh"
#include "Thread/Lock.hh"
#include "Thread/LockStatistics.hh"

#ifndef INCLUDED_ERRNO_H
#include "Exceptions/Errno.hh"
#endif

#include <pthread.h>
#include <sys/time.h>

namespace fatalmind {

template<class MT>
template<class TM>
bool CondVar<MT>::wait(const struct POSIX::timespec<TM>* timeout) {
	struct ::timeval now;
	gettimeofday(&now, NULL);

	POSIX::timespec<SingleThreadedModel<NopLockStatistics> > ts(now.tv_sec, now.tv_usec*1000);
	ts+= *timeout;
	return _waitAbsolute(&ts);
}

template<class MT>
bool CondVar<MT>::_waitAbsolute(const struct POSIX::timespec<SingleThreadedModel<NopLockStatistics> >* timeout) {
	super::stat_unlock();
	try {
		Errno::ThrowOnError(
			pthread_cond_timedwait(&d_condvar, &d_mutex, timeout)
		);
	} catch (Errno::ETIMEDOUT e) {
		super::stat_lock();
		return false;
	}
	super::stat_lock();
	return true;
}

template bool CondVar<PosixMutex<> >::wait(const struct POSIX::timespec<SingleThreadedModel<NopLockStatistics> >* timeout);
template bool CondVar<PosixMutex<> >::wait(const struct POSIX::timespec<MultiThreadedModel<NopLockStatistics> >* timeout);

template bool CondVar<PosixMutex<DefaultPosixMutexSpecialization, UsageOnlyLockStatistics> >::wait(const struct POSIX::timespec<SingleThreadedModel<NopLockStatistics> >* timeout);
template bool CondVar<PosixMutex<DefaultPosixMutexSpecialization, UsageOnlyLockStatistics> >::wait(const struct POSIX::timespec<MultiThreadedModel<NopLockStatistics> >* timeout);

} // namespace fatalmind
