//******************************************************************** 
//*** t/main.c                                                     
//*** Copyright (c) 2001 by Markus Winand <mws@fatalmind.com>       
//*** $Id: main.cc,v 1.33 2005/03/18 22:34:33 mws Exp $     
//******************************************************************** 

#include <stdio.h>
#include <string>
#include <sys/time.h>
#include <sys/resource.h>
#include <iostream>
#include <sstream>
#include <iomanip>
#include <unistd.h>
#include <memory>
#include <stdexcept>


#include "../Exceptions/Errno.hh"
#include "../POSIX/sys/timeval.hh"
#include "main.hh"
#include "ResourcePool/SQL/SQLException.hh"

using namespace std;
using namespace fatalmind;

int OK = 0, FAIL = 0, ERROR = 0;

void _assertNotPass(char* file, int line)
{
	ostringstream os;
	os << "assertNotPass failed in " << file << ":" << line;
	throw fatalmind::Exception(os.str());
}


void _assertTrue(bool b, char* s, char* file, int line) {
	if (! b) {
		ostringstream os;
		os << s << " failed in " << file << ":" << line;
		throw fatalmind::Exception(os.str());
	}
}

		tester::tester(int s, int l)
		: FAIL(0)
		, silent(s)
		, loop(l) 
		{
		};

		int tester::run_tests() {
			FAIL = 0;
			num  = 0;
			OK   = 0;
			normal_test();
			return FAIL;
		}
		int tester::run_allperf() {
			FAIL = 0;
			num  = 0;
			OK   = 0;
			run_perf();
			return FAIL;
		}

		void tester::test_it(void func(), char *name) {
			bool result(false);
			string errstr;
			
			try {
				func();
				result = true;
			} catch(string s) {
				errstr = s;
			} catch (SQL::SQLException e) {
				errstr = "SQLException: " + e.message() + " with SQL '" + e.getSQL() + "'";
			} catch(Errno eno) {
				errstr = "ERRNO: ";
				errstr += eno.message();
			} catch(Exception ex) {
				errstr = "Exceptions::Exception: " + ex.message();
			} catch(std::exception ex){
				errstr = "std::exception: ";
				errstr.append(ex.what());
			} catch(...) {
				errstr = "Unkown excpetion";
			}
			
			if (! silent || ! result) {
				cout << std::setw(3) << std::right << num 
					<< " " << std::left << std::setw(67) << name;
			}

			if (result) {  
				if (! silent) {
					cout << "  OK" << std::endl;
				}
				OK++;
			} else {
				cout << "FAIL" << std::endl;
				FAIL++;
				if (errstr.length()) {
					cout << " ---> " << errstr << std::endl;
				}
			}

			num++;
		}

		void tester::printSummary() {
			cout << "===" << endl;
			cout << "all                                                     OK: " << std::setw(3) << OK << "  FAIL: " << FAIL << std::endl;
		}

int run_tests();
int run_perf();

void print_rusage(struct rusage& rusg) {
	cout << rusg.ru_utime.tv_sec << '.' << rusg.ru_utime.tv_usec << "    " 
		<< rusg.ru_stime.tv_sec << '.' << rusg.ru_stime.tv_usec << endl;
}

long dump_rusagediff_usr(struct rusage& rusg_start, struct rusage& rusg_stop) {
	long user;
	user = rusg_stop.ru_utime.tv_sec-rusg_start.ru_utime.tv_sec;
	user*=10000;
	user += (rusg_stop.ru_utime.tv_usec-rusg_start.ru_utime.tv_usec)/100;
	return user;
}

long dump_rusagediff_sys(struct rusage& rusg_start, struct rusage& rusg_stop) {
	long sys;
	sys = rusg_stop.ru_stime.tv_sec-rusg_start.ru_stime.tv_sec;
	sys*=100;
	sys += (rusg_stop.ru_stime.tv_usec-rusg_start.ru_stime.tv_usec)/100;
	return sys;
}

long cpu_diff(struct rusage& rusg_start, struct rusage& rusg_stop) {
	return dump_rusagediff_sys(rusg_start, rusg_stop) +
		dump_rusagediff_usr(rusg_start, rusg_stop);
}

auto_ptr<tester> getTester(int, int);

	int silent=0;
	int loop=0;
	int listoutput = 0;

void testrunit(tester* tc, int no, int perf_seconds = 16) {
		struct rusage rusg_start, rusg_stop;
		struct rusage rusg_ch_start, rusg_ch_stop;
		int i;
		long user, sys;
		int loops = 10;
		float rt = 0;
		int factor = perf_seconds;
		POSIX::timeval rt_start;	
		POSIX::timeval rt_stop;	
		struct timezone tz;

		gettimeofday(&rt_start, &tz);
		i = 0;
		do {
			tc->run_single_perf(no);
			gettimeofday(&rt_stop, &tz);
			++i;
			rt = (rt_stop - rt_start);
		} while ((rt < 1) && (tc->getFAIL() == 0));
		if (rt > 1.050) {
			factor = (int)((1/rt) * perf_seconds);
		}
		loops = i*factor;
		loops = loops?loops:1;
		//loops *= 8; // for long testing

		//printf("DB> will do loops %d (calibration is %d when rt was %f, factor =%d)\n", loops, i, rt, factor);

		gettimeofday(&rt_start, &tz);
		getrusage(RUSAGE_SELF, &rusg_start);
		getrusage(RUSAGE_CHILDREN, &rusg_ch_start);
		
		for (i = 0; (i < loops) && (tc->getFAIL() == 0); i++) {
			tc->run_single_perf(no);
		}	
		getrusage(RUSAGE_SELF, &rusg_stop);
		getrusage(RUSAGE_CHILDREN, &rusg_ch_stop);
		gettimeofday(&rt_stop, &tz);

		user  =  dump_rusagediff_usr(rusg_start   , rusg_stop);
		user +=  dump_rusagediff_usr(rusg_ch_start, rusg_ch_stop);
		sys   =  dump_rusagediff_sys(rusg_start   , rusg_stop);
		sys  +=  dump_rusagediff_sys(rusg_ch_start, rusg_ch_stop);
		float cost_p_l = (float)(user+sys)/(float)i/10.0;
		float rtime_p_l = ((float)(rt_stop-rt_start)*1000)/i;
		if (! listoutput) {
			printf("\nUsage statistics: (after %d testruns) for %s\n", i, tc->get_single_perf_name(no));
			printf("=================\n");
			printf("  Real time: %7.4f s\n", (float)(rt_stop-rt_start));
			printf("  User time: %7.4f s\n", (float)user/(float)10000);
			printf("  Sys  time: %7.4f s\n", (float)sys/(float)10000);
			printf("\n");
			printf("   cost per loop: %-6.3f ms \n", cost_p_l);
			printf("  rtime per loop: %-6.3f ms \n", rtime_p_l);
			printf("   wait per loop: %-6.3f ms \n", rtime_p_l - cost_p_l);
			printf("      wait ratio: %5.2f%% \n", (rtime_p_l - cost_p_l)*100.0/rtime_p_l);
		} else {
			cout << std::setw(40) << tc->get_single_perf_name(no)
				<< std::setw(5) << std::fixed << std::showpoint << i
				<< std::setw(8) << std::setprecision(2) << (float)(rt_stop-rt_start)
				<< std::setw(8) << std::setprecision(4) << (float)user/(float)10000
				<< std::setw(8) << std::setprecision(4) << (float)sys/(float)10000
				<< std::setw(10) << std::setprecision(3) << cost_p_l
				<< std::setw(10) << std::setprecision(3) << rtime_p_l
				<< std::setw(10) << std::setprecision(3) << rtime_p_l - cost_p_l
				<< std::setw(8) << std::setprecision(2) << ((rtime_p_l - cost_p_l)*100.0)/rtime_p_l << '%' 
				<< std::endl;
		}
}

void printlistheader() {
	cout << std::setw(40) << "Test name"
		<< std::setw(5) << "i"
		<< std::setw(8) << "real"
		<< std::setw(8) << "usr"
		<< std::setw(8) << "sys"
		<< std::setw(10) << "cost/l"
		<< std::setw(10) << "RT/l"
		<< std::setw(10) << "wait/l"
		<< std::setw(8) << "wait%"
		<< std::endl;
}

extern char *optarg;
extern int optind, opterr, optopt;

int main(int argc, char*argv[]) {
	int test_fail = 0;
	int perf_seconds = 16;
	
	char ch;
	while ((ch = getopt(argc, argv, "spl")) != -1) {
		switch (ch) {
			case 's': silent = 1; break;
			case 'p': silent = 1; loop++; break;
			case 'l': listoutput = 1; printlistheader(); break;
		};
	}
	if (optind < argc) {
		if (loop >= 2) {
			sscanf(argv[optind], "%d", &perf_seconds);	
		} else {
			std::cout << "additional params " << argc-optind << " ignored " << std::endl;
		}
	}

	auto_ptr<tester> tc(getTester(silent, loop));
	tc->init();
	test_fail = tc->run_tests();
	if (test_fail && loop) {
		printf("some tests failed\n");
		printf("fix them before makeing performance tests\n");
	}
	if (loop && ! test_fail) {
		if (loop == 1) {
			testrunit(tc.get(), 0);		
		} else {
			for (int i = 1; i <= tc->getSinglePerfTests(); ++i) {
				testrunit(tc.get(), i, perf_seconds);
			}
		}
		tc->cleanup();
	} else {
		tc->cleanup();
		tc->printSummary();
	}
	exit(tc->getFAIL());
}

